from typing import TypedDict, NotRequired

# enabling type hints for misc stuff


class ConnectionParams(TypedDict):
    """
    - *dbname*: the database name
    - *database*: the database name (only as keyword argument)
    - *user*: user name used to authenticate
    - *password*: password used to authenticate
    - *host*: database host address (defaults to UNIX socket if not provided)
    - *port*: connection port number (defaults to 5432 if not provided)
    """
    pass
